#include <scene_rdl2/scene/rdl2/rdl2.h>
using namespace scene_rdl2::rdl2;
RDL2_DSO_ATTR_DECLARE_NS(ImageNormalMap_attr)
AttributeKey<String> attrTexture;
AttributeKey<Bool> attrUseDefaultValue;
AttributeKey<Vec3f> attrDefaultValue;
AttributeKey<Bool> attrWrapAround;
AttributeKey<Int> attrNormalEncoding;
AttributeKey<Int> attrTextureEnum;
AttributeKey<Vec3f> attrInputTextureCoordinate;
AttributeKey<Vec2f> attrOffset;
AttributeKey<Vec2f> attrScale;
AttributeKey<Float> attrRotationAngle;
AttributeKey<Vec2f> attrRotationCenter;
RDL2_DSO_ATTR_DEFINE(NormalMap)
ImageNormalMap_attr::attrTexture = sceneClass.declareAttribute<String>("tangent_space_normal_texture", FLAGS_FILENAME, INTERFACE_GENERIC, {});
sceneClass.setMetadata(ImageNormalMap_attr::attrTexture, "comment", "Filename that points to a tangent space normal texture .exr or .tx file (must be mip-mapped and tiled with maketx).   If 'udim_files' is empty and a 'UDIM' token is present in the filename, then 'UDIM' will be replaced with the udim number (i.e. 1001, 1002, etc.)");
ImageNormalMap_attr::attrUseDefaultValue = sceneClass.declareAttribute<Bool>("use_default_value_when_missing", false, FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setMetadata(ImageNormalMap_attr::attrUseDefaultValue, "comment", "Uses the 'default_value' for missing udims and does not report error");
ImageNormalMap_attr::attrDefaultValue = sceneClass.declareAttribute<Vec3f>("default_value", Vec3f(0.0f, 0.0f, 1.0f), FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setMetadata(ImageNormalMap_attr::attrDefaultValue, "comment", "Default value to be used for missing udims when 'use_default_value_when_missing' is enabled");
sceneClass.setMetadata(ImageNormalMap_attr::attrDefaultValue, "enable if", "OrderedDict({'use_default_value_when_missing': 'true'})");
ImageNormalMap_attr::attrWrapAround = sceneClass.declareAttribute<Bool>("wrap_around", true, FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setMetadata(ImageNormalMap_attr::attrWrapAround, "label", "wrap around");
sceneClass.setMetadata(ImageNormalMap_attr::attrWrapAround, "comment", "Controls whether to repeat (true) or clamp (false) the texture");
ImageNormalMap_attr::attrNormalEncoding = sceneClass.declareAttribute<Int>("normal_encoding", 0, FLAGS_ENUMERABLE, INTERFACE_GENERIC, {});
sceneClass.setEnumValue(ImageNormalMap_attr::attrNormalEncoding, 0, "[0,1]");
sceneClass.setEnumValue(ImageNormalMap_attr::attrNormalEncoding, 1, "[-1,1]");
sceneClass.setMetadata(ImageNormalMap_attr::attrNormalEncoding, "label", "normal encoding");
sceneClass.setMetadata(ImageNormalMap_attr::attrNormalEncoding, "comment", "Most normal maps are encoded [0,1]. Only certain rare floating point normal maps are encoded [-1,1]");
ImageNormalMap_attr::attrTextureEnum = sceneClass.declareAttribute<Int>("texture_coordinates", 0, FLAGS_ENUMERABLE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrTextureEnum);
sceneClass.setEnumValue(ImageNormalMap_attr::attrTextureEnum, 0, "texture");
sceneClass.setEnumValue(ImageNormalMap_attr::attrTextureEnum, 1, "input texture coordinates");
sceneClass.setMetadata(ImageNormalMap_attr::attrTextureEnum, "label", "texture coordinates");
sceneClass.setMetadata(ImageNormalMap_attr::attrTextureEnum, "comment", "Selects which texture coordinates to use.   'texture' uses the surface UVs.   'input texture coordinates' uses the map bound to 'input_texture_coordinates'.");
ImageNormalMap_attr::attrInputTextureCoordinate = sceneClass.declareAttribute<Vec3f>("input_texture_coordinates", Vec3f(0.0f, 0.0f, 0.0f), FLAGS_BINDABLE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrInputTextureCoordinate);
sceneClass.setMetadata(ImageNormalMap_attr::attrInputTextureCoordinate, "label", "input texture coordinates");
sceneClass.setMetadata(ImageNormalMap_attr::attrInputTextureCoordinate, "enable if", "OrderedDict({'texture_coordinates': '1'})");
sceneClass.setMetadata(ImageNormalMap_attr::attrInputTextureCoordinate, "comment", "When 'texture_coordinates' is set to 'input texture coordinates', UVs are generated by evaluating the map bound to this attribute");
ImageNormalMap_attr::attrOffset = sceneClass.declareAttribute<Vec2f>("offset", Vec2f(0.0f, 0.0f), FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrOffset);
sceneClass.setMetadata(ImageNormalMap_attr::attrOffset, "comment", "Translates the UV coordinates in texture space.  The order of operations is rotation, scale, and then offset.");
ImageNormalMap_attr::attrScale = sceneClass.declareAttribute<Vec2f>("scale", Vec2f(1.0f, 1.0f), FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrScale);
sceneClass.setMetadata(ImageNormalMap_attr::attrScale, "comment", "Scales the UV coordinates in texture space.  The order of operations is rotation, scale, and then offset.");
ImageNormalMap_attr::attrRotationAngle = sceneClass.declareAttribute<Float>("rotation_angle", 0.0f, FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrRotationAngle);
sceneClass.setMetadata(ImageNormalMap_attr::attrRotationAngle, "label", "rotation angle");
sceneClass.setMetadata(ImageNormalMap_attr::attrRotationAngle, "comment", "Rotation of the UV coordinates in texture space in degrees.  The order of operations is rotation, scale, and then offset.");
ImageNormalMap_attr::attrRotationCenter = sceneClass.declareAttribute<Vec2f>("rotation_center", Vec2f(0.5f, 0.5f), FLAGS_NONE, INTERFACE_GENERIC, {});
sceneClass.setGroup("UVs", ImageNormalMap_attr::attrRotationCenter);
sceneClass.setMetadata(ImageNormalMap_attr::attrRotationCenter, "label", "rotation center");
sceneClass.setMetadata(ImageNormalMap_attr::attrRotationCenter, "comment", "The UV coordinate around which to rotate.  The order of operations is rotation, scale, and then offset.");
RDL2_DSO_ATTR_END
using namespace ImageNormalMap_attr;
